import pandas as pd 


def cols_to_strip_commas(df, columns):
    """Strip commas from selected dataframe columns.

    Args: 
        df: Pandas dataframe.
        columns: List of columns to convert.

    Returns: 
        Original dataframe with converted column data.
    """

    for col in columns:
        df[col] = df[col].str.replace(',','')
        
    return df


def cols_to_drop(df, columns):
    """Drop selected columns and return dataframe.

    Args: 
        df: Pandas dataframe.
        columns: List of columns to drop.

    Returns: 
        Original dataframe without dropped columns.
    """

    for col in columns:
        df.drop([col], axis=1, inplace=True) 

    return df


def col_names_to_lower(df):
    """Convert column names to lowercase and return dataframe.

    Args: 
        df: Pandas dataframe.

    Returns: 
        Original dataframe with converted column data.
    """

    return df.columns.str.lower()


def cols_to_rename(df, dictionary):
    """Drop selected columns and return dataframe.
    
    Args: 
        df: Pandas dataframe.
        dictionary: {'old_name1': 'new_name1', 'old_name2': 'new_name2'}

    Returns: 
        Original dataframe without dropped columns.
    """

    df.rename(dictionary, axis=1, inplace=True)

    return df


import pandas as pd
import numpy as np


def cols_to_log(df, columns):
    """Transform column data with log and return new columns of prefixed data.

    For us with data where the column values do not include zeroes.

    Args: 
        df: Pandas dataframe.
        columns: List of columns to transform.

    Returns: 
        Original dataframe with additional prefixed columns.
    """

    for col in columns:
        df['log_'+col] = np.log(df[col])

    return df


def cols_to_log1p(df, columns):
    """Transform column data with log+1 and return new columns of prefixed data. 
    
    For use with data where the column values include zeroes. 

    Args: 
        df: Pandas dataframe.
        columns: List of columns to transform.

    Returns: 
        Original dataframe with additional prefixed columns.
    """

    for col in columns:
        df['log1p_'+col] = np.log(df[col]+1)

    return df


def cols_to_log_max_root(df, columns):
    """Convert data points to log values using the maximum value as the log max and return new columns of prefixed data. 
    
    For use with data where the column values include zeroes. 

    Args: 
        df: Pandas dataframe.
        columns: List of columns to transform.

    Returns: 
        Original dataframe with additional prefixed columns.
    """

    for col in columns:
        log_max = np.log(df[col].max())
        df['logmr_'+col] = df[col] ** (1 / log_max)

    return df


def cols_to_tanh(df, columns):
    """Transform column data with hyperbolic tangent and return new columns of prefixed data. 
    
    Args: 
        df: Pandas dataframe.
        columns: List of columns to transform.

    Returns: 
        Original dataframe with additional prefixed columns.
    """

    for col in columns:
        df['tanh_'+col] = np.tanh(df[col])

    return df


def cols_to_sigmoid(df, columns):
    """Convert data points to values between 0 and 1 using a sigmoid function and return new columns of prefixed data. 
    
    Args: 
        df: Pandas dataframe.
        columns: List of columns to transform.

    Returns: 
        Original dataframe with additional prefixed columns.
    """
    
    for col in columns:
        e = np.exp(1)
        y = 1 / (1+e**(-df[col]))
        df['sig_'+col] = y

    return df


def cols_to_cube_root(df, columns):
    """Convert data points to their cube root value so all values are between 0-1 and return new columns of prefixed data. 
    
    Args: 
        df: Pandas dataframe.
        columns: List of columns to transform.

    Returns: 
        Original dataframe with additional prefixed columns.
    """
    
    for col in columns:
        df['cube_root_'+col] = df[col] ** (1/3)

    return df


def cols_to_cube_root_normalize(df, columns):
    """Convert data points to their normalized cube root value so all values are between 0-1 and return new columns of prefixed data. 
    
    Args: 
        df: Pandas dataframe.
        columns: List of columns to transform.

    Returns: 
        Original dataframe with additional prefixed columns.
    """
    
    for col in columns:
        df['cube_root_'+col] = (df[col] - df[col].min()) / (df[col].max() - df[col].min()) ** (1/3)

    return df

def cols_to_percentile(df, columns):
    """Convert data points to their percentile linearized value and return new columns of prefixed data. 
    
    Args: 
        df: Pandas dataframe.
        columns: List of columns to transform.

    Returns: 
        Original dataframe with additional prefixed columns.
    """
    for col in columns:
        df['pc_lin_'+col] = df[col].rank(method='min').apply(lambda x: (x-1) / len(df[col])-1)

    return df


def cols_to_normalize(df, columns):
    """Convert data points to values between 0 and 1 and return new columns of prefixed data.
    
    Args: 
        df: Pandas dataframe.
        columns: List of columns to transform.

    Returns: 
        Original dataframe with additional prefixed columns.
    """
    
    for col in columns:
        df['norm_'+col] = (df[col] - df[col].min()) / (df[col].max() - df[col].min())

    return df


def cols_to_log1p_normalize(df, columns):
    """Transform column data with log+1 normalized and return new columns of prefixed data. 
    
    For use with data where the column values include zeroes. 

    Args: 
        df: Pandas dataframe.
        columns: List of columns to transform.

    Returns: 
        Original dataframe with additional prefixed columns.
    """

    for col in columns:
        df['log1p_norm_'+col] = np.log((df[col] - df[col].min()) / (df[col].max() - df[col].min())+1)

    return df


def cols_to_one_hot(df, columns):
    """One hot encode column values and return new prefixed columns.
    
    Args: 
        df: Pandas dataframe.
        columns: List of columns to transform.

    Returns: 
        Original dataframe with additional prefixed columns.
    """
    
    for col in columns:
        encoding = pd.get_dummies(df[col], prefix=col)
        df = pd.concat([df, encoding], axis=1)
        
    return df


def cols_to_reduce_uniques(df, column_threshold_dict):
    """Reduce the number of unique values by creating a column of X values and the rest marked "Others".
    
    Args: 
        df: Pandas dataframe.
        columns: Dictionary of column and threshold, i.e. {'col1' : 1000, 'col2' : 3000}

    Returns: 
        Original dataframe with additional prefixed columns. The most dominant values in the column will
        be assigned their original value. The less dominant results will be assigned to Others, which can
        help visualise and model data in some cases.
    """
    
    for key, value in column_threshold_dict.items():
            counts = df[key].value_counts()
            others = set(counts[counts < value].index)
            df['reduce_'+key] = df[key].replace(list(others), 'Others')
            
    return df


def get_grouped_stats(df, group, column):
    """Group by a column and return summary statistics for a given column in new columns.

    Args:
        df: Pandas dataframe.
        group: Column name to groupby
        column: Column to summarise.

    Returns:
        Original dataframe with new column containing previous value of named column.

    """

    df['mean_' + column] = df.groupby([group])[column].transform('mean')
    df['median_' + column] = df.groupby([group])[column].transform('median')
    df['std_' + column] = df.groupby([group])[column].transform('std')
    df['max_' + column] = df.groupby([group])[column].transform('max')
    df['min_' + column] = df.groupby([group])[column].transform('min')

    return df


def get_previous_value(df, group, column, name):
    """Group by a column and return the previous value of another column and assign value to a new column.

    Args:
        df: Pandas dataframe.
        group: Column name to groupby
        column: Column value to return.
        name: Name for new column.

    Returns:
        Original dataframe with new column containing previous value of named column.

    """
    df[name] = df.groupby([group])[column].shift().notnull().astype(int)
    return df


def get_dates(df, date_column):
    """Converts a given date to various formats and returns an updated dataframe.

    Args:
        df: Pandas dataframe.
        columns: List of columns to transform.

    Returns:
        Original dataframe with additional date columns.
    """

    df['day'] = df[date_column].dt.strftime("%d")  # Day of month with leading zero
    df['month'] = df[date_column].dt.strftime("%m")  # Month of year with leading zero
    df['year'] = df[date_column].dt.strftime("%Y")  # Full numeric four digit year
    df['year_month'] = df[date_column].dt.strftime("%Y%m")  # Full numeric four digit year plus month
    df['week_number'] = df[date_column].dt.strftime("%U")  # Week number with leading zero
    df['day_number'] = df[date_column].dt.strftime("%j")  # Day number with leading zero
    df['day_name'] = df[date_column].dt.strftime("%A")  # Day name, i.e. Sunday
    df['month_name'] = df[date_column].dt.strftime("%B")  # Month name, i.e. January
    df['mysql_date'] = df[date_column].dt.strftime("%Y-%d-%m")  # MySQL date, i.e. 2020-30-01

    return df




import pandas as pd


def cols_to_slugify(df, columns):
    """Slugify selected column values and return dataframe.

    Args: 
        df: Pandas dataframe.
        columns: List of columns to convert.

    Returns: 
        Original dataframe with converted column data.
    """

    for col in columns:
        df[col] = df[col].str.replace('[^0-9a-zA-Z]+', '_', regex=True)

    return df


def cols_to_float(df, columns):
    """Convert selected column values to float and return dataframe.

    Args: 
        df: Pandas dataframe.
        columns: List of columns to convert.

    Returns: 
        Original dataframe with converted column data.
    """

    for col in columns:
        df[col] = df[col].astype(float)

    return df


def cols_to_int(df, columns):
    """Convert selected column values to int and return dataframe.

    Args: 
        df: Pandas dataframe.
        columns: List of columns to convert.

    Returns: 
        Original dataframe with converted column data.
    """

    for col in columns:
        df[col] = df[col].astype(int)

    return df


def cols_to_datetime(df, columns):
    """Convert selected column values to datetime and return dataframe.

    Args: 
        df: Pandas dataframe.
        columns: List of columns to convert.

    Returns: 
        Original dataframe with converted column data.
    """

    for col in columns:
        df[col] = pd.to_datetime(df[col])

    return df


def cols_to_negative(df, columns):
    """Convert selected column values to negative and return dataframe.

    Args: 
        df: Pandas dataframe.
        columns: List of columns to convert.

    Returns: 
        Original dataframe with converted column data.
    """

    for col in columns:
        df[col] = df[col] * -1

    return df


def num_to_Cat(df, column, labels, bins):
    """

    :return:
    """
    try:
        df[column] = pd.cut(df[column], bins=bins, labels=labels)
    except Exception as e:
        print("Unexpected error :", e)
        raise ValueError()

    return df


def cat_to_cat(df, column, dict):
    """
     non-exhaustive mapping : retains the existing variables for non-matches,  add fillna

    :param df:
    :param column:
    :param dict:
    :return:
    """
    try:
        df[column] = df[column].map(dict)
    except Exception as e:
        print("Unexpected error :", e)
        raise ValueError()

    return df




#General functions

def checkNull(df, columnName, acceptableAmount):

    """

     # Run QA checks to confirm data population
    isNameValid = checkNull(df,'full_name',.9)
    isAgeValid = checkNull(df,'age',.9)
    isGenderValid = checkNull(df,'gender',1)

    :param df:
    :param columnName:
    :param acceptableAmount:
    :return:
    """
    acceptableQuality = False
    nullPercent = df[columnName].isnull().sum() / df[columnName].count()
    if nullPercent >= acceptableAmount:
        acceptableQuality = True

    return acceptableQuality


def swap_col(df,A,B):
    """

    :param df:
    :param A:
    :param B:
    :return:
    """
    df.loc[:, [B, A]] = df[[A, B]].to_numpy()

    return df

#Generic Functions for validation

def valid_date_format(date):
    """Validates a general date by making sure it is 8 digits long, are only digits and the year can only contain a 2 or 1 in the front (to exlude 0123,3000,4000)

    Args:
        date:str
    Returns:
        returns Boolean
    """
    if (len(date)==8 and date.isdigit() and (date[0]=="2" or date[0]=="1")):
        return True
    else:
        return False

def no_spaces(field):
    """Validates that there are no spaces in a string

    Args:
        field:str
    Returns:
        returns Boolean
    """
    if (' ' in field) == True:
        return(False)
    else:
        return(True)


def title_case(field):
    """Validates that as sting is title case
    Args:
        field:str
    Returns:
        returns Boolean
    """
    if (field.istitle()):
        return(True)
    else:
        return(False)


def is_greater_than(field,notgreaterthanthis):
    """Validates that a field length is not greater than the notgreaterthanthis value.

    Args:
        field:str
        notgreaterthanthis: int
    Returns:
        returns Boolean
    """
    if len(field) > notgreaterthanthis:
        return(False)
    else:
        return(True)
#Generic Functions for Normalisation

def normalize_date(date):
        """Removes non digits and joins numbers together

        Args:
            date:str
        Returns:
            returns str
        """
        whitelist = set('0123456789')
        normalized_date = ''.join(filter(whitelist.__contains__, date))
        return(str(normalized_date))

def remove_spaces(string):
        """Removes spaces

        Args:
            string:str
        Returns:
            returns str
        """
        string = string.replace(" ", "")
        return str(string)

def title_case(string):
         """makes string title case

         Args:
             string:str
         Returns:
             returns str
         """
         string = string.lower().title()
         return string

def remove_suffix(string):
          """Removes suffixes

          Args:
              string:str
          Returns:
              returns str
          """
          suffixes = ["Esq", "Ii", "Iii", "Iiii", "Iv", "Jnr", "Jr", "Sr"]
          string = string.replace(" ", "")
          string = string.replace(".", "")
          string = string.replace(",", "")
          for suffix in suffixes:
                  if(string.endswith(suffix)):
                          string = string[:-len(suffix)]
                          return(string)
          return(string)


#Generic Functions

def normalize_date(date):
        """Removes non digits and joins numbers together

        Args:
            date:str
        Returns:
            returns str
        """
        whitelist = set('0123456789')
        normalized_date = ''.join(filter(whitelist.__contains__, date))
        return(str(normalized_date))

def remove_spaces(string):
        """Removes spaces

        Args:
            string:str
        Returns:
            returns str
        """
        string = string.replace(" ", "")
        return str(string)

def title_case(string):
         """makes string title case

         Args:
             string:str
         Returns:
             returns str
         """
         string = string.lower().title()
         return string

def remove_suffix(string):
          """Removes suffixes

          Args:
              string:str
          Returns:
              returns str
          """
          suffixes = ["Esq", "Ii", "Iii", "Iiii", "Iv", "Jnr", "Jr", "Sr"]
          string = string.replace(" ", "")
          string = string.replace(".", "")
          string = string.replace(",", "")
          for suffix in suffixes:
                  if(string.endswith(suffix)):
                          string = string[:-len(suffix)]
                          return(string)
          return(string)
      
def say_hi():
    return "Hi Data Engineer"